#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, 2011, Robert Collins <robertc@robertcollins.net>
# 
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.


"""Included fixtures."""

__all__ = [
    'ByteStream',
    'DetailStream',
    'EnvironmentVariable',
    'EnvironmentVariableFixture',
    'FakeLogger',
    'FakePopen',
    'LoggerFixture',
    'LogHandler',
    'MockPatch',
    'MockPatchMultiple',
    'MockPatchObject',
    'MonkeyPatch',
    'NestedTempfile',
    'PackagePathEntry',
    'PopenFixture',
    'PythonPackage',
    'PythonPathEntry',
    'StringStream',
    'TempDir',
    'TempHomeDir',
    'Timeout',
    'TimeoutException',
    'WarningsCapture',
    'WarningsFilter',
]


from fixtures._fixtures.environ import (
    EnvironmentVariable,
    EnvironmentVariableFixture,
    )
from fixtures._fixtures.logger import (
    FakeLogger,
    LoggerFixture,
    LogHandler,
    )
from fixtures._fixtures.mockpatch import (
    MockPatch,
    MockPatchMultiple,
    MockPatchObject,
    )
from fixtures._fixtures.monkeypatch import MonkeyPatch
from fixtures._fixtures.popen import (
    FakePopen,
    PopenFixture,
    )
from fixtures._fixtures.packagepath import PackagePathEntry
from fixtures._fixtures.pythonpackage import PythonPackage
from fixtures._fixtures.pythonpath import PythonPathEntry
from fixtures._fixtures.streams import (
    ByteStream,
    DetailStream,
    StringStream,
    )
from fixtures._fixtures.tempdir import (
    NestedTempfile,
    TempDir,
    )
from fixtures._fixtures.temphomedir import (
    TempHomeDir,
    )
from fixtures._fixtures.timeout import (
    Timeout,
    TimeoutException,
    )
from fixtures._fixtures.warnings import (
    WarningsCapture,
    WarningsFilter,
    )



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, 2011, Robert Collins <robertc@robertcollins.net>
# 
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    'FakePopen',
    'PopenFixture'
    ]

import random
import subprocess
import sys

from fixtures import Fixture


class FakeProcess(object):
    """A test double process, roughly meeting subprocess.Popen's contract."""

    def __init__(self, args, info):
        self._args = args
        self.stdin = info.get('stdin')
        self.stdout = info.get('stdout')
        self.stderr = info.get('stderr')
        self.pid = random.randint(0, 65536)
        self._returncode = info.get('returncode', 0)
        self.returncode = None

    @property
    def args(self):
        return self._args["args"]

    def poll(self):
        """Get the current value of FakeProcess.returncode.

        The returncode is None before communicate() and/or wait() are called,
        and it's set to the value provided by the 'info' dictionary otherwise
        (or 0 in case 'info' doesn't specify a value).
        """
        return self.returncode

    def communicate(self, input=None, timeout=None):
        self.returncode = self._returncode
        if self.stdin and input:
            self.stdin.write(input)
        if self.stdout:
            out = self.stdout.getvalue()
        else:
            out = ''
        if self.stderr:
            err = self.stderr.getvalue()
        else:
            err = ''
        return out, err

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.wait()

    def kill(self):
        pass

    def wait(self, timeout=None, endtime=None):
        if self.returncode is None:
            self.communicate()
        return self.returncode


class FakePopen(Fixture):
    """Replace subprocess.Popen.

    Primarily useful for testing, this fixture replaces subprocess.Popen with a
    test double.

    :ivar procs: A list of the processes created by the fixture.
    """

    _unpassed = object()

    def __init__(self, get_info=lambda _:{}):
        """Create a PopenFixture

        :param get_info: Optional callback to control the behaviour of the
            created process. This callback takes a kwargs dict for the Popen
            call, and should return a dict with any desired attributes.
            Only parameters that are supplied to the Popen call are in the
            dict, making it possible to detect the difference between 'passed
            with a default value' and 'not passed at all'.

            e.g. 
            def get_info(proc_args):
                self.assertEqual(subprocess.PIPE, proc_args['stdin'])
                return {'stdin': StringIO('foobar')}

            The default behaviour if no get_info is supplied is for the return
            process to have returncode of None, empty streams and a random pid.

            After communicate() or wait() are called on the process object,
            the returncode is set to whatever get_info returns (or 0 if
            get_info is not supplied or doesn't return a dict with an explicit
            'returncode' key).
        """
        super(FakePopen, self).__init__()
        self.get_info = get_info

    def _setUp(self):
        self.addCleanup(setattr, subprocess, 'Popen', subprocess.Popen)
        subprocess.Popen = self
        self.procs = []

    # The method has the correct signature so we error appropriately if called
    # wrongly.
    def __call__(self, args, bufsize=_unpassed, executable=_unpassed,
        stdin=_unpassed, stdout=_unpassed, stderr=_unpassed,
        preexec_fn=_unpassed, close_fds=_unpassed, shell=_unpassed,
        cwd=_unpassed, env=_unpassed, universal_newlines=_unpassed,
        startupinfo=_unpassed, creationflags=_unpassed,
        restore_signals=_unpassed, start_new_session=_unpassed,
        pass_fds=_unpassed, *, group=_unpassed, extra_groups=_unpassed,
        user=_unpassed, umask=_unpassed, encoding=_unpassed,
        errors=_unpassed, text=_unpassed, pipesize=_unpassed,
        process_group=_unpassed):
        # Reject arguments introduced by newer versions of Python in older
        # versions; this makes it harder to accidentally hide compatibility
        # problems using test doubles.
        if sys.version_info < (3, 7) and text is not FakePopen._unpassed:
            raise TypeError(
                "FakePopen.__call__() got an unexpected keyword argument "
                "'text'")
        if sys.version_info < (3, 9):
            for arg_name in "group", "extra_groups", "user", "umask":
                if locals()[arg_name] is not FakePopen._unpassed:
                    raise TypeError(
                        "FakePopen.__call__() got an unexpected keyword "
                        "argument '{}'".format(arg_name))
        if sys.version_info < (3, 10) and pipesize is not FakePopen._unpassed:
            raise TypeError(
                "FakePopen.__call__() got an unexpected keyword argument "
                "'pipesize'")
        if sys.version_info < (3, 11) and process_group is not FakePopen._unpassed:
            raise TypeError(
                "FakePopen.__call__() got an unexpected keyword argument "
                "'process_group'")

        proc_args = dict(args=args)
        local = locals()
        for param in [
            "bufsize", "executable", "stdin", "stdout", "stderr",
            "preexec_fn", "close_fds", "shell", "cwd", "env",
            "universal_newlines", "startupinfo", "creationflags",
            "restore_signals", "start_new_session", "pass_fds", "group",
            "extra_groups", "user", "umask", "encoding", "errors", "text",
            "pipesize", "process_group"]:
            if local[param] is not FakePopen._unpassed:
                proc_args[param] = local[param]
        proc_info = self.get_info(proc_args)
        result = FakeProcess(proc_args, proc_info)
        self.procs.append(result)
        return result


PopenFixture = FakePopen



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from unittest import mock # Yes, we only test the rolling backport
import testtools

from fixtures import (
    MockPatch,
    MockPatchMultiple,
    MockPatchObject,
)


class Foo(object):
    def bar(self):
        return self


def mocking_bar(self):
    return 'mocked!'


class TestMockPatch(testtools.TestCase):
    def test_mock_patch_with_replacement(self):
        self.useFixture(MockPatch('%s.Foo.bar' % (__name__), mocking_bar))
        instance = Foo()
        self.assertEqual(instance.bar(), 'mocked!')

    def test_mock_patch_without_replacement(self):
        self.useFixture(MockPatch('%s.Foo.bar' % (__name__)))
        instance = Foo()
        self.assertIsInstance(instance.bar(), mock.MagicMock)


class TestMockMultiple(testtools.TestCase):
    def test_mock_multiple_with_replacement(self):
        self.useFixture(MockPatchMultiple('%s.Foo' % (__name__),
                                          bar=mocking_bar))
        instance = Foo()
        self.assertEqual(instance.bar(), 'mocked!')

    def test_mock_patch_without_replacement(self):
        self.useFixture(MockPatchMultiple(
            '%s.Foo' % (__name__),
            bar=MockPatchMultiple.DEFAULT))
        instance = Foo()
        self.assertIsInstance(instance.bar(), mock.MagicMock)


class TestMockPatchObject(testtools.TestCase):
    def test_mock_patch_object_with_replacement(self):
        self.useFixture(MockPatchObject(Foo, 'bar', mocking_bar))
        instance = Foo()
        self.assertEqual(instance.bar(), 'mocked!')

    def test_mock_patch_object_without_replacement(self):
        self.useFixture(MockPatchObject(Foo, 'bar'))
        instance = Foo()
        self.assertIsInstance(instance.bar(), mock.MagicMock)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2011, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import io
import logging
import time

import testtools
from testtools import TestCase

from fixtures import (
    FakeLogger,
    LogHandler,
    TestWithFixtures,
    )


# A simple custom formatter that prepends Foo to all log messages, for
# testing formatter overrides.
class FooFormatter(logging.Formatter):
    def format(self, record):
        self._style = logging.PercentStyle("Foo " + self._style._fmt)
        self._fmt = self._style._fmt
        return logging.Formatter.format(self, record)


class FakeLoggerTest(TestCase, TestWithFixtures):

    def setUp(self):
        super(FakeLoggerTest, self).setUp()
        self.logger = logging.getLogger()
        self.addCleanup(self.removeHandlers, self.logger)

    def removeHandlers(self, logger):
        for handler in logger.handlers:
            logger.removeHandler(handler)

    def test_output_property_has_output(self):
        fixture = self.useFixture(FakeLogger())
        logging.info("some message")
        self.assertEqual("some message\n", fixture.output)

    def test_replace_and_restore_handlers(self):
        stream = io.StringIO()
        logger = logging.getLogger()
        logger.addHandler(logging.StreamHandler(stream))
        logger.setLevel(logging.INFO)
        logging.info("one")
        fixture = FakeLogger()
        with fixture:
            logging.info("two")
        logging.info("three")
        self.assertEqual("two\n", fixture.output)
        self.assertEqual("one\nthree\n", stream.getvalue())

    def test_preserving_existing_handlers(self):
        stream = io.StringIO()
        self.logger.addHandler(logging.StreamHandler(stream))
        self.logger.setLevel(logging.INFO)
        fixture = FakeLogger(nuke_handlers=False)
        with fixture:
            logging.info("message")
        self.assertEqual("message\n", fixture.output)
        self.assertEqual("message\n", stream.getvalue())

    def test_logging_level_restored(self):
        self.logger.setLevel(logging.DEBUG)
        fixture = FakeLogger(level=logging.WARNING)
        with fixture:
            # The fixture won't capture this, because the DEBUG level
            # is lower than the WARNING one
            logging.debug("debug message")
            self.assertEqual(logging.WARNING, self.logger.level)
        self.assertEqual("", fixture.output)
        self.assertEqual(logging.DEBUG, self.logger.level)

    def test_custom_format(self):
        fixture = FakeLogger(format="%(module)s")
        self.useFixture(fixture)
        logging.info("message")
        self.assertEqual("test_logger\n", fixture.output)

    def test_custom_datefmt(self):
        fixture = FakeLogger(format="%(asctime)s %(module)s",
                             datefmt="%Y")
        self.useFixture(fixture)
        logging.info("message")
        self.assertEqual(
            time.strftime("%Y test_logger\n", time.localtime()),
            fixture.output)

    def test_custom_formatter(self):
        fixture = FakeLogger(format="%(asctime)s %(module)s",
                             formatter=FooFormatter,
                             datefmt="%Y")
        self.useFixture(fixture)
        logging.info("message")
        self.assertEqual(
            time.strftime("Foo %Y test_logger\n", time.localtime()),
            fixture.output)

    def test_logging_output_included_in_details(self):
        fixture = FakeLogger()
        detail_name = "pythonlogging:''"
        with fixture:
            content = fixture.getDetails()[detail_name]
            # Output after getDetails is called is included.
            logging.info('some message')
            self.assertEqual("some message\n", content.as_text())
        # The old content object returns the old usage after cleanUp (not
        # strictly needed but convenient). Note that no guarantee is made that
        # it will work after setUp is called again. [It does on Python 2.x, not
        # on 3.x]
        self.assertEqual("some message\n", content.as_text())
        with fixture:
            # A new one returns new output:
            self.assertEqual("", fixture.getDetails()[detail_name].as_text())
        # The original content object may either fail, or return the old
        # content (it must not have been reset..).
        try:
            self.assertEqual("some message\n", content.as_text())
        except AssertionError:
            raise
        except:
            pass

    def test_exceptionraised(self):
        with FakeLogger():
            with testtools.ExpectedException(TypeError):
                logging.info("Some message", "wrongarg")

    def test_output_can_be_reset(self):
        fixture = FakeLogger()
        with fixture:
            logging.info("message")

        fixture.reset_output()

        self.assertEqual("", fixture.output)


class LogHandlerTest(TestCase, TestWithFixtures):

    class CustomHandler(logging.Handler):

        def __init__(self, *args, **kwargs):
            """Create the instance, and add a records attribute."""
            logging.Handler.__init__(self, *args, **kwargs)
            self.msgs = []

        def emit(self, record):
            self.msgs.append(record.msg)

    def setUp(self):
        super(LogHandlerTest, self).setUp()
        self.logger = logging.getLogger()
        self.addCleanup(self.removeHandlers, self.logger)

    def removeHandlers(self, logger):
        for handler in logger.handlers:
            logger.removeHandler(handler)

    def test_captures_logging(self):
        fixture = self.useFixture(LogHandler(self.CustomHandler()))
        logging.info("some message")
        self.assertEqual(["some message"], fixture.handler.msgs)

    def test_replace_and_restore_handlers(self):
        stream = io.StringIO()
        logger = logging.getLogger()
        logger.addHandler(logging.StreamHandler(stream))
        logger.setLevel(logging.INFO)
        logging.info("one")
        fixture = LogHandler(self.CustomHandler())
        with fixture:
            logging.info("two")
        logging.info("three")
        self.assertEqual(["two"], fixture.handler.msgs)
        self.assertEqual("one\nthree\n", stream.getvalue())

    def test_preserving_existing_handlers(self):
        stream = io.StringIO()
        self.logger.addHandler(logging.StreamHandler(stream))
        self.logger.setLevel(logging.INFO)
        fixture = LogHandler(self.CustomHandler(), nuke_handlers=False)
        with fixture:
            logging.info("message")
        self.assertEqual(["message"], fixture.handler.msgs)
        self.assertEqual("message\n", stream.getvalue())

    def test_logging_level_restored(self):
        self.logger.setLevel(logging.DEBUG)
        fixture = LogHandler(self.CustomHandler(), level=logging.WARNING)
        with fixture:
            # The fixture won't capture this, because the DEBUG level
            # is lower than the WARNING one
            logging.debug("debug message")
            self.assertEqual(logging.WARNING, self.logger.level)
        self.assertEqual([], fixture.handler.msgs)
        self.assertEqual(logging.DEBUG, self.logger.level)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, 2011, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import inspect
import io
import subprocess
import sys

import testtools

from fixtures import FakePopen, TestWithFixtures
from fixtures._fixtures.popen import FakeProcess


class TestFakePopen(testtools.TestCase, TestWithFixtures):

    def test_installs_restores_global(self):
        fixture = FakePopen()
        popen = subprocess.Popen
        fixture.setUp()
        try:
            self.assertEqual(subprocess.Popen, fixture)
        finally:
            fixture.cleanUp()
            self.assertEqual(subprocess.Popen, popen)

    def test___call___is_recorded(self):
        fixture = self.useFixture(FakePopen())
        proc = fixture(['foo', 'bar'], 1, None, 'in', 'out', 'err')
        self.assertEqual(1, len(fixture.procs))
        self.assertEqual(
            dict(
                args=['foo', 'bar'], bufsize=1, executable=None,
                stdin='in', stdout='out', stderr='err',
            ),
            proc._args,
        )

    def test_inject_content_stdout(self):
        def get_info(args):
            return {'stdout': 'stdout'}
        fixture = self.useFixture(FakePopen(get_info))
        proc = fixture(['foo'])
        self.assertEqual('stdout', proc.stdout)

    def test_handles_all_Popen_args(self):
        all_args = dict(
            args="args", bufsize="bufsize", executable="executable",
            stdin="stdin", stdout="stdout", stderr="stderr",
            preexec_fn="preexec_fn", close_fds="close_fds", shell="shell",
            cwd="cwd", env="env", universal_newlines="universal_newlines",
            startupinfo="startupinfo", creationflags="creationflags",
            restore_signals="restore_signals",
            start_new_session="start_new_session", pass_fds="pass_fds",
            encoding="encoding", errors="errors")
        if sys.version_info >= (3, 7):
            all_args["text"] = "text"
        if sys.version_info >= (3, 9):
            all_args["group"] = "group"
            all_args["extra_groups"] = "extra_groups"
            all_args["user"] = "user"
            all_args["umask"] = "umask"
        if sys.version_info >= (3, 10):
            all_args["pipesize"] = "pipesize"
        if sys.version_info >= (3, 11):
            all_args["process_group"] = "process_group"

        def get_info(proc_args):
            self.assertEqual(all_args, proc_args)
            return {}

        fixture = self.useFixture(FakePopen(get_info))
        fixture(**all_args)

    @testtools.skipUnless(
        sys.version_info < (3, 7), "only relevant on Python <3.7")
    def test_rejects_3_7_args_on_older_versions(self):
        fixture = self.useFixture(FakePopen(lambda proc_args: {}))
        with testtools.ExpectedException(
                TypeError, r".* got an unexpected keyword argument 'text'"):
            fixture(args="args", text=True)

    @testtools.skipUnless(
        sys.version_info < (3, 9), "only relevant on Python <3.9")
    def test_rejects_3_9_args_on_older_versions(self):
        fixture = self.useFixture(FakePopen(lambda proc_args: {}))
        for arg_name in ("group", "extra_groups", "user", "umask"):
            kwargs = {arg_name: arg_name}
            expected_message = (
                r".* got an unexpected keyword argument '{}'".format(arg_name))
            with testtools.ExpectedException(TypeError, expected_message):
                fixture(args="args", **kwargs)

    @testtools.skipUnless(
        sys.version_info < (3, 10), "only relevant on Python <3.10")
    def test_rejects_3_10_args_on_older_versions(self):
        fixture = self.useFixture(FakePopen(lambda proc_args: {}))
        with testtools.ExpectedException(
                TypeError,
                r".* got an unexpected keyword argument 'pipesize'"):
            fixture(args="args", pipesize=1024)

    @testtools.skipUnless(
        sys.version_info < (3, 11), "only relevant on Python <3.11")
    def test_rejects_3_11_args_on_older_versions(self):
        fixture = self.useFixture(FakePopen(lambda proc_args: {}))
        with testtools.ExpectedException(
                TypeError,
                r".* got an unexpected keyword argument 'process_group'"):
            fixture(args="args", process_group=42)

    def test_function_signature(self):
        fake_signature = inspect.getfullargspec(FakePopen.__call__)
        real_signature = inspect.getfullargspec(subprocess.Popen)

        # compare args

        fake_args = fake_signature.args
        real_args = real_signature.args

        self.assertListEqual(
            fake_args,
            real_args,
            "Function signature of FakePopen doesn't match subprocess.Popen",
        )

        # compare kw-only args; order doesn't matter so we use a set

        fake_kwargs = set(fake_signature.kwonlyargs)
        real_kwargs = set(real_signature.kwonlyargs)

        if sys.version_info < (3, 11):
            fake_kwargs.remove('process_group')

        if sys.version_info < (3, 10):
            fake_kwargs.remove('pipesize')

        if sys.version_info < (3, 9):
            fake_kwargs.remove('group')
            fake_kwargs.remove('extra_groups')
            fake_kwargs.remove('user')
            fake_kwargs.remove('umask')

        if sys.version_info < (3, 7):
            fake_kwargs.remove('text')

        self.assertSetEqual(
            fake_kwargs,
            real_kwargs,
            "Function signature of FakePopen doesn't match subprocess.Popen",
        )

    def test_custom_returncode(self):
        def get_info(proc_args):
            return dict(returncode=1)
        proc = self.useFixture(FakePopen(get_info))(['foo'])
        self.assertEqual(None, proc.returncode)
        self.assertEqual(1, proc.wait())
        self.assertEqual(1, proc.returncode)

    def test_with_popen_custom(self):
        self.useFixture(FakePopen())
        with subprocess.Popen(['ls -lh']) as proc:
            self.assertEqual(None, proc.returncode)
            self.assertEqual(['ls -lh'], proc.args)


class TestFakeProcess(testtools.TestCase):

    def test_wait(self):
        proc = FakeProcess({}, {})
        proc.returncode = 45
        self.assertEqual(45, proc.wait())

    def test_communicate(self):
        proc = FakeProcess({}, {})
        self.assertEqual(('', ''), proc.communicate())
        self.assertEqual(0, proc.returncode)

    def test_communicate_with_out(self):
        proc = FakeProcess({}, {'stdout': io.BytesIO(b'foo')})
        self.assertEqual((b'foo', ''), proc.communicate())
        self.assertEqual(0, proc.returncode)

    def test_communicate_with_input(self):
        proc = FakeProcess({}, {'stdout': io.BytesIO(b'foo')})
        self.assertEqual((b'foo', ''), proc.communicate(input=b'bar'))

    def test_communicate_with_input_and_stdin(self):
        stdin = io.BytesIO()
        proc = FakeProcess({}, {'stdin': stdin})
        proc.communicate(input=b'hello')
        self.assertEqual(b'hello', stdin.getvalue())

    def test_communicate_with_timeout(self):
        proc = FakeProcess({}, {'stdout': io.BytesIO(b'foo')})
        self.assertEqual((b'foo', ''), proc.communicate(timeout=10))

    def test_args(self):
        proc = FakeProcess({"args": ["ls", "-lh"]}, {})
        proc.returncode = 45
        self.assertEqual(45, proc.wait())
        self.assertEqual(proc.args, ["ls", "-lh"])

    def test_kill(self):
        proc = FakeProcess({}, {})
        self.assertIs(None, proc.kill())

    def test_poll(self):
        proc = FakeProcess({}, {})
        self.assertIs(None, proc.poll())
        proc.communicate()
        self.assertEqual(0, proc.poll())

    def test_poll_with_returncode(self):
        proc = FakeProcess({}, {})
        proc.communicate()
        self.assertEqual(0, proc.poll())

    def test_wait_with_timeout_and_endtime(self):
        proc = FakeProcess({}, {})
        self.assertEqual(0, proc.wait(timeout=4, endtime=7))



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
# 
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import unittest
import testtools
from testtools.content import text_content
from testtools.testcase import skipIf

import fixtures
from fixtures import TestWithFixtures
from fixtures.fixture import gather_details
from fixtures.tests.helpers import LoggingFixture


class TestTestWithFixtures(unittest.TestCase):

    def test_useFixture(self):
        fixture = LoggingFixture()
        class SimpleTest(testtools.TestCase, TestWithFixtures):
            def test_foo(self):
                self.useFixture(fixture)
        result = unittest.TestResult()
        SimpleTest('test_foo').run(result)
        self.assertTrue(result.wasSuccessful())
        self.assertEqual(['setUp', 'cleanUp'], fixture.calls)

    def test_useFixture_uses_raise_first(self):
        calls = []
        def raiser(ignored):
            calls.append('called')
            raise Exception('foo')
        fixture = fixtures.FunctionFixture(lambda:None, raiser)
        class SimpleTest(testtools.TestCase, TestWithFixtures):
            def test_foo(self):
                self.useFixture(fixture)
        result = unittest.TestResult()
        SimpleTest('test_foo').run(result)
        self.assertFalse(result.wasSuccessful())
        self.assertEqual(['called'], calls)

    @skipIf(gather_details is None, "gather_details() is not available.")
    def test_useFixture_details_captured_from_setUp(self):
        # Details added during fixture set-up are gathered even if setUp()
        # fails with an exception.
        class SomethingBroke(Exception): pass
        class BrokenFixture(fixtures.Fixture):
            def setUp(self):
                super(BrokenFixture, self).setUp()
                self.addDetail('content', text_content("foobar"))
                raise SomethingBroke()
        broken_fixture = BrokenFixture()
        class DetailedTestCase(TestWithFixtures, testtools.TestCase):
            def setUp(self):
                super(DetailedTestCase, self).setUp()
                self.useFixture(broken_fixture)
            def test(self):
                pass
        detailed_test_case = DetailedTestCase("test")
        self.assertRaises(SomethingBroke, detailed_test_case.setUp)
        self.assertEqual(
            {"content": text_content("foobar")},
            broken_fixture.getDetails())
        self.assertEqual(
            {"content": text_content("foobar")},
            detailed_test_case.getDetails())

    @skipIf(gather_details is None, "gather_details() is not available.")
    def test_useFixture_details_not_captured_from_setUp(self):
        # Details added during fixture set-up are not gathered if the test
        # case does not have the ability to accept those details.
        class SomethingBroke(Exception): pass
        class BrokenFixture(fixtures.Fixture):
            def setUp(self):
                super(BrokenFixture, self).setUp()
                self.addDetail('content', text_content("foobar"))
                raise SomethingBroke()
        broken_fixture = BrokenFixture()
        class NonDetailedTestCase(TestWithFixtures, unittest.TestCase):
            def setUp(self):
                super(NonDetailedTestCase, self).setUp()
                self.useFixture(broken_fixture)
            def test(self):
                pass
        non_detailed_test_case = NonDetailedTestCase("test")
        self.assertRaises(SomethingBroke, non_detailed_test_case.setUp)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import types

import testtools
from testtools.content import text_content
from testtools.testcase import skipIf

import fixtures
from fixtures.fixture import gather_details
from fixtures.tests.helpers import LoggingFixture


require_gather_details = skipIf(gather_details is None,
        "gather_details() is not available.")


# Note: the cleanup related tests are strictly speaking redundant, IFF they are
# replaced with contract tests for correct use of CallMany.
class TestFixture(testtools.TestCase):

    def test_resetCallsSetUpCleanUp(self):
        calls = []
        class FixtureWithSetupOnly(fixtures.Fixture):
            def setUp(self):
                super(FixtureWithSetupOnly, self).setUp()
                calls.append('setUp')
                self.addCleanup(calls.append, 'cleanUp')
        fixture = FixtureWithSetupOnly()
        fixture.setUp()
        fixture.reset()
        fixture.cleanUp()
        self.assertEqual(['setUp', 'cleanUp', 'setUp', 'cleanUp'], calls)

    def test_reset_raises_if_cleanup_raises(self):
        class FixtureWithSetupOnly(fixtures.Fixture):
            def do_raise(self):
                raise Exception('foo')
            def setUp(self):
                super(FixtureWithSetupOnly, self).setUp()
                self.addCleanup(self.do_raise)
        fixture = FixtureWithSetupOnly()
        fixture.setUp()
        exc = self.assertRaises(Exception, fixture.reset)
        self.assertEqual(('foo',), exc.args)

    def test_cleanUp_raise_first_false_callscleanups_returns_exceptions(self):
        calls = []
        def raise_exception1():
            calls.append('1')
            raise Exception('woo')
        def raise_exception2():
            calls.append('2')
            raise Exception('woo')
        class FixtureWithException(fixtures.Fixture):
            def setUp(self):
                super(FixtureWithException, self).setUp()
                self.addCleanup(raise_exception2)
                self.addCleanup(raise_exception1)
        fixture = FixtureWithException()
        fixture.setUp()
        exceptions = fixture.cleanUp(raise_first=False)
        self.assertEqual(['1', '2'], calls)
        # There should be two exceptions
        self.assertEqual(2, len(exceptions))
        # They should be a sys.exc_info tuple.
        self.assertEqual(3, len(exceptions[0]))
        type, value, tb = exceptions[0]
        self.assertEqual(Exception, type)
        self.assertIsInstance(value, Exception)
        self.assertEqual(('woo',), value.args)
        self.assertIsInstance(tb, types.TracebackType)

    def test_exit_propagates_exceptions(self):
        fixture = fixtures.Fixture()
        fixture.__enter__()
        self.assertEqual(False, fixture.__exit__(None, None, None))

    def test_exit_runs_all_raises_first_exception(self):
        calls = []
        def raise_exception1():
            calls.append('1')
            raise Exception('woo')
        def raise_exception2():
            calls.append('2')
            raise Exception('hoo')
        class FixtureWithException(fixtures.Fixture):
            def setUp(self):
                super(FixtureWithException, self).setUp()
                self.addCleanup(raise_exception2)
                self.addCleanup(raise_exception1)
        fixture = FixtureWithException()
        fixture.__enter__()
        exc = self.assertRaises(Exception, fixture.__exit__, None, None, None)
        self.assertEqual(('woo',), exc.args[0][1].args)
        self.assertEqual(('hoo',), exc.args[1][1].args)
        self.assertEqual(['1', '2'], calls)

    def test_useFixture(self):
        parent = LoggingFixture('-outer')
        nested = LoggingFixture('-inner', calls=parent.calls)
        parent.setUp()
        parent.useFixture(nested)
        parent.cleanUp()
        self.assertEqual(
            ['setUp-outer', 'setUp-inner', 'cleanUp-inner', 'cleanUp-outer'],
            parent.calls)

    @require_gather_details
    def test_useFixture_details_captured_from_setUp(self):
        # Details added during fixture set-up are gathered even if setUp()
        # fails with an unknown exception.
        class SomethingBroke(Exception): pass
        class BrokenFixture(fixtures.Fixture):
            def setUp(self):
                super(BrokenFixture, self).setUp()
                self.addDetail('content', text_content("foobar"))
                raise SomethingBroke()
        broken_fixture = BrokenFixture()
        class SimpleFixture(fixtures.Fixture):
            def setUp(self):
                super(SimpleFixture, self).setUp()
                self.useFixture(broken_fixture)
        simple_fixture = SimpleFixture()
        self.assertRaises(SomethingBroke, simple_fixture.setUp)
        self.assertEqual(
            {"content": text_content("foobar")},
            broken_fixture.getDetails())
        self.assertEqual(
            {"content": text_content("foobar")},
            simple_fixture.getDetails())

    @require_gather_details
    def test_useFixture_details_captured_from_setUp_MultipleExceptions(self):
        # Details added during fixture set-up are gathered even if setUp()
        # fails with (cleanly - with MultipleExceptions / SetupError).
        class SomethingBroke(Exception): pass
        class BrokenFixture(fixtures.Fixture):
            def _setUp(self):
                self.addDetail('content', text_content("foobar"))
                raise SomethingBroke()
        class SimpleFixture(fixtures.Fixture):
            def _setUp(self):
                self.useFixture(BrokenFixture())
        simple = SimpleFixture()
        e = self.assertRaises(fixtures.MultipleExceptions, simple.setUp)
        self.assertEqual(
            {"content": text_content("foobar")},
            e.args[-1][1].args[0])

    def test_getDetails(self):
        fixture = fixtures.Fixture()
        with fixture:
            self.assertEqual({}, fixture.getDetails())

    def test_details_from_child_fixtures_are_returned(self):
        parent = fixtures.Fixture()
        with parent:
            child = fixtures.Fixture()
            parent.useFixture(child)
            # Note that we add the detail *after* using the fixture: the parent
            # has to query just-in-time.
            child.addDetail('foo', 'content')
            self.assertEqual({'foo': 'content'}, parent.getDetails())
            # And dropping it from the child drops it from the parent.
            del child._details['foo']
            self.assertEqual({}, parent.getDetails())
            # After cleanup the child details are still gone.
            child.addDetail('foo', 'content')
        self.assertRaises(TypeError, parent.getDetails)

    def test_duplicate_details_are_disambiguated(self):
        parent = fixtures.Fixture()
        with parent:
            parent.addDetail('foo', 'parent-content')
            child = fixtures.Fixture()
            parent.useFixture(child)
            # Note that we add the detail *after* using the fixture: the parent
            # has to query just-in-time.
            child.addDetail('foo', 'child-content')
            self.assertEqual({'foo': 'parent-content',
                              'foo-1': 'child-content',}, parent.getDetails())

    def test_addDetail(self):
        fixture = fixtures.Fixture()
        with fixture:
            fixture.addDetail('foo', 'content')
            self.assertEqual({'foo': 'content'}, fixture.getDetails())
            del fixture._details['foo']
            self.assertEqual({}, fixture.getDetails())
            fixture.addDetail('foo', 'content')
        # Cleanup clears the details too.
        self.assertRaises(TypeError, fixture.getDetails)

    def test_setUp_subclassed(self):
        # Even though its no longer recommended, we need to be sure that
        # overriding setUp and calling super().setUp still works.
        class Subclass(fixtures.Fixture):
            def setUp(self):
                super(Subclass, self).setUp()
                self.fred = 1
                self.addCleanup(setattr, self, 'fred', 2)
        with Subclass() as f:
            self.assertEqual(1, f.fred)
        self.assertEqual(2, f.fred)

    def test__setUp(self):
        # _setUp is called, and cleanups can be registered by it.
        class Subclass(fixtures.Fixture):
            def _setUp(self):
                self.fred = 1
                self.addCleanup(setattr, self, 'fred', 2)
        with Subclass() as f:
            self.assertEqual(1, f.fred)
        self.assertEqual(2, f.fred)

    def test__setUp_fails(self):
        # when _setUp fails, the fixture is left ready-to-setUp, and any
        # details added during _setUp are captured.
        class Subclass(fixtures.Fixture):
            def _setUp(self):
                self.addDetail('log', text_content('stuff'))
                1/0
        f = Subclass()
        e = self.assertRaises(fixtures.MultipleExceptions, f.setUp)
        self.assertRaises(TypeError, f.cleanUp)
        self.assertIsInstance(e.args[0][1], ZeroDivisionError)
        self.assertIsInstance(e.args[1][1], fixtures.SetupError)
        self.assertEqual('stuff', e.args[1][1].args[0]['log'].as_text())

    def test__setUp_fails_cleanUp_fails(self):
        # when _setUp fails, cleanups are called, and their failure is captured
        # into the MultipleExceptions instance.
        class Subclass(fixtures.Fixture):
            def _setUp(self):
                self.addDetail('log', text_content('stuff'))
                self.addCleanup(lambda: 1/0)
                raise Exception('fred')
        f = Subclass()
        e = self.assertRaises(fixtures.MultipleExceptions, f.setUp)
        self.assertRaises(TypeError, f.cleanUp)
        self.assertEqual(Exception, e.args[0][0])
        self.assertEqual(ZeroDivisionError, e.args[1][0])
        self.assertEqual(fixtures.SetupError, e.args[2][0])
        self.assertEqual('stuff', e.args[2][1].args[0]['log'].as_text())

    def test_setup_failures_with_base_exception(self):
        # when _setUp fails with a BaseException (or subclass thereof) that
        # exception is propagated as is, but we still call cleanups etc.
        class MyBase(BaseException):pass
        log = []
        class Subclass(fixtures.Fixture):
            def _setUp(self):
                self.addDetail('log', text_content('stuff'))
                self.addCleanup(log.append, 'cleaned')
                raise MyBase('fred')
        f = Subclass()
        self.assertRaises(MyBase, f.setUp)
        self.assertRaises(TypeError, f.cleanUp)
        self.assertEqual(['cleaned'], log)


class TestFunctionFixture(testtools.TestCase):

    def test_setup_only(self):
        fixture = fixtures.FunctionFixture(lambda: 42)
        fixture.setUp()
        self.assertEqual(42, fixture.fn_result)
        fixture.cleanUp()
        self.assertFalse(hasattr(fixture, 'fn_result'))

    def test_cleanup(self):
        results = []
        fixture = fixtures.FunctionFixture(lambda: 84, results.append)
        fixture.setUp()
        self.assertEqual(84, fixture.fn_result)
        self.assertEqual([], results)
        fixture.cleanUp()
        self.assertEqual([84], results)

    def test_reset(self):
        results = []
        expected = [21, 7]
        def setUp():
            return expected.pop(0)
        def reset(result):
            results.append(('reset', result))
            return expected.pop(0)
        fixture = fixtures.FunctionFixture(setUp, results.append, reset)
        fixture.setUp()
        self.assertEqual([], results)
        fixture.reset()
        self.assertEqual([('reset', 21)], results)
        self.assertEqual(7, fixture.fn_result)
        fixture.cleanUp()
        self.assertEqual([('reset', 21), 7], results)


class TestMethodFixture(testtools.TestCase):

    def test_no_setup_cleanup(self):
        class Stub:
            pass
        fixture = fixtures.MethodFixture(Stub())
        fixture.setUp()
        fixture.reset()
        self.assertIsInstance(fixture.obj, Stub)
        fixture.cleanUp()

    def test_setup_only(self):
        class Stub:
            def setUp(self):
                self.value = 42
        fixture = fixtures.MethodFixture(Stub())
        fixture.setUp()
        self.assertEqual(42, fixture.obj.value)
        self.assertIsInstance(fixture.obj, Stub)
        fixture.cleanUp()

    def test_cleanup_only(self):
        class Stub:
            value = None
            def tearDown(self):
                self.value = 42
        fixture = fixtures.MethodFixture(Stub())
        fixture.setUp()
        self.assertEqual(None, fixture.obj.value)
        self.assertIsInstance(fixture.obj, Stub)
        fixture.cleanUp()
        self.assertEqual(42, fixture.obj.value)

    def test_cleanup(self):
        class Stub:
            def setUp(self):
                self.value = 42
            def tearDown(self):
                self.value = 84
        fixture = fixtures.MethodFixture(Stub())
        fixture.setUp()
        self.assertEqual(42, fixture.obj.value)
        self.assertIsInstance(fixture.obj, Stub)
        fixture.cleanUp()
        self.assertEqual(84, fixture.obj.value)

    def test_custom_setUp(self):
        class Stub:
            def mysetup(self):
                self.value = 42
        obj = Stub()
        fixture = fixtures.MethodFixture(obj, setup=obj.mysetup)
        fixture.setUp()
        self.assertEqual(42, fixture.obj.value)
        self.assertEqual(obj, fixture.obj)
        fixture.cleanUp()

    def test_custom_cleanUp(self):
        class Stub:
            value = 42
            def mycleanup(self):
                self.value = None
        obj = Stub()
        fixture = fixtures.MethodFixture(obj, cleanup=obj.mycleanup)
        fixture.setUp()
        self.assertEqual(42, fixture.obj.value)
        self.assertEqual(obj, fixture.obj)
        fixture.cleanUp()
        self.assertEqual(None, fixture.obj.value)

    def test_reset(self):
        class Stub:
            def setUp(self):
                self.value = 42
            def tearDown(self):
                self.value = 84
            def reset(self):
                self.value = 126
        obj = Stub()
        fixture = fixtures.MethodFixture(obj, reset=obj.reset)
        fixture.setUp()
        self.assertEqual(obj, fixture.obj)
        self.assertEqual(42, obj.value)
        fixture.reset()
        self.assertEqual(126, obj.value)
        fixture.cleanUp()
        self.assertEqual(84, obj.value)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
# 
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import fixtures

class LoggingFixture(fixtures.Fixture):

    def __init__(self, suffix='', calls=None):
        super(LoggingFixture, self).__init__()
        if calls is None:
            calls = []
        self.calls = calls
        self.suffix = suffix

    def setUp(self):
        super(LoggingFixture, self).setUp()
        self.calls.append('setUp' + self.suffix)
        self.addCleanup(self.calls.append, 'cleanUp' + self.suffix)

    def reset(self):
        self.calls.append('reset' + self.suffix)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import os


def load_tests(loader, standard_tests, pattern):
    test_modules = [
        os.path.splitext(path)[0]
        for path in os.listdir(os.path.dirname(__file__))
        if path.startswith('test_')
    ]
    prefix = "fixtures.tests._fixtures."
    test_mod_names = [prefix + test_module for test_module in test_modules]
    standard_tests.addTests(loader.loadTestsFromNames(test_mod_names))
    return standard_tests



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
# 
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import functools
import sys

import testtools
from testtools.matchers import Is

from fixtures import MonkeyPatch, TestWithFixtures

reference = 23

NEW_PY39_CLASSMETHOD = (
    sys.version_info[:2] in ((3, 9), (3,10))
    and not hasattr(sys, "pypy_version_info"))

class C(object):
    def foo(self, arg):
        return arg
    @staticmethod
    def foo_static(): pass
    @classmethod
    def foo_cls(cls): pass

class D(object):
    def bar(self): pass
    def bar_two_args(self, arg=None):
        return (self, arg)
    @classmethod
    def bar_cls(cls):
        return cls
    @classmethod
    def bar_cls_args(cls, *args):
        return (cls,) + args
    @staticmethod
    def bar_static():
        pass
    @staticmethod
    def bar_static_args(*args):
        return args
    def bar_self_referential(self, *args, **kwargs):
        self.bar()

INST_C = C()

def fake(*args):
    return args
def fake2(arg): pass
def fake_no_args(): pass
def fake_no_args2(): pass
@staticmethod
def fake_static(): pass


class TestMonkeyPatch(testtools.TestCase, TestWithFixtures):

    def test_patch_and_restore(self):
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.reference', 45)
        self.assertEqual(23, reference)
        fixture.setUp()
        try:
            self.assertEqual(45, reference)
        finally:
            fixture.cleanUp()
            self.assertEqual(23, reference)

    def test_patch_missing_attribute(self):
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.new_attr', True)
        self.assertFalse('new_attr' in globals())
        fixture.setUp()
        try:
            self.assertEqual(True, new_attr)
        finally:
            fixture.cleanUp()
            self.assertFalse('new_attr' in globals())

    def test_delete_existing_attribute(self):
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.reference',
            MonkeyPatch.delete)
        self.assertEqual(23, reference)
        fixture.setUp()
        try:
            self.assertFalse('reference' in globals())
        finally:
            fixture.cleanUp()
            self.assertEqual(23, reference)

    def test_delete_missing_attribute(self):
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.new_attr',
            MonkeyPatch.delete)
        self.assertFalse('new_attr' in globals())
        fixture.setUp()
        try:
            self.assertFalse('new_attr' in globals())
        finally:
            fixture.cleanUp()
            self.assertFalse('new_attr' in globals())

    def _check_restored_static_or_class_method(self, oldmethod, oldmethod_inst,
            klass, methodname):
        restored_method = getattr(klass, methodname)
        restored_method_inst = getattr(klass(), methodname)
        self.assertEqual(oldmethod, restored_method)
        self.assertEqual(oldmethod, restored_method_inst)
        self.assertEqual(oldmethod_inst, restored_method)
        self.assertEqual(oldmethod_inst, restored_method_inst)
        restored_method()
        restored_method_inst()

    def test_patch_staticmethod_with_staticmethod(self):
        oldmethod = C.foo_static
        oldmethod_inst = C().foo_static
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_static',
            D.bar_static)
        with fixture:
            C.foo_static()
            C().foo_static()
        self._check_restored_static_or_class_method(oldmethod, oldmethod_inst,
                C, 'foo_static')

    def test_patch_staticmethod_with_classmethod(self):
        oldmethod = C.foo_static
        oldmethod_inst = C().foo_static
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_static',
            D.bar_cls)
        with fixture:
            C.foo_static()
            C().foo_static()
        self._check_restored_static_or_class_method(oldmethod, oldmethod_inst,
                C, 'foo_static')

    def test_patch_staticmethod_with_function(self):
        oldmethod = C.foo_static
        oldmethod_inst = C().foo_static
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_static',
            fake_no_args)
        with fixture:
            C.foo_static()
            C().foo_static()
        self._check_restored_static_or_class_method(oldmethod, oldmethod_inst,
                C, 'foo_static')

    def test_patch_staticmethod_with_boundmethod(self):
        oldmethod = C.foo_static
        oldmethod_inst = C().foo_static
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_static',
            D().bar)
        with fixture:
            C.foo_static()
            C().foo_static()
        self._check_restored_static_or_class_method(oldmethod, oldmethod_inst,
                C, 'foo_static')

    def test_patch_classmethod_with_staticmethod(self):
        oldmethod = C.foo_cls
        oldmethod_inst = C().foo_cls
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_cls',
            D.bar_static_args)
        with fixture:
            (cls,) = C.foo_cls()
            self.expectThat(cls, Is(C))
            (cls,) = C().foo_cls()
            self.expectThat(cls, Is(C))
        self._check_restored_static_or_class_method(oldmethod, oldmethod_inst,
                C, 'foo_cls')

    def test_patch_classmethod_with_classmethod(self):
        oldmethod = C.foo_cls
        oldmethod_inst = C().foo_cls
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_cls',
            D.bar_cls_args)
        with fixture:
            # Python 3.9 changes the behavior of the classmethod decorator so
            # that it now honours the descriptor binding protocol [1].
            # This means we're now going to call the monkeypatched classmethod
            # the way we would have if it hadn't been monkeypatched: simply
            # with the class
            #
            # https://bugs.python.org/issue19072
            if NEW_PY39_CLASSMETHOD:
                cls, = C.foo_cls()
                self.expectThat(cls, Is(D))
                cls, = C().foo_cls()
                self.expectThat(cls, Is(D))
            else:
                cls, target_class = C.foo_cls()
                self.expectThat(cls, Is(D))
                self.expectThat(target_class, Is(C))
                cls, target_class = C().foo_cls()
                self.expectThat(cls, Is(D))
                self.expectThat(target_class, Is(C))
        self._check_restored_static_or_class_method(
            oldmethod, oldmethod_inst, C, 'foo_cls')

    def test_patch_classmethod_with_function(self):
        oldmethod = C.foo_cls
        oldmethod_inst = C().foo_cls
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_cls',
            fake)
        with fixture:
            (cls,) = C.foo_cls()
            self.expectThat(cls, Is(C))
            (cls, arg) = C().foo_cls(1)
            self.expectThat(cls, Is(C))
            self.assertEqual(1, arg)
        self._check_restored_static_or_class_method(oldmethod, oldmethod_inst,
                C, 'foo_cls')

    def test_patch_classmethod_with_boundmethod(self):
        oldmethod = C.foo_cls
        oldmethod_inst = C().foo_cls
        d = D()
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_cls',
            d.bar_two_args)
        with fixture:
            slf, cls = C.foo_cls()
            self.expectThat(slf, Is(d))
            # See note in test_patch_classmethod_with_classmethod on changes in
            # Python 3.9
            if NEW_PY39_CLASSMETHOD:
                self.expectThat(cls, Is(None))
            else:
                self.expectThat(cls, Is(C))
            slf, cls = C().foo_cls()
            self.expectThat(slf, Is(d))
            if NEW_PY39_CLASSMETHOD:
                self.expectThat(cls, Is(None))
            else:
                self.expectThat(cls, Is(C))
        self._check_restored_static_or_class_method(
            oldmethod, oldmethod_inst, C, 'foo_cls')

    def test_patch_function_with_staticmethod(self):
        oldmethod = fake_no_args
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.fake_no_args',
            D.bar_static)
        with fixture:
            fake_no_args()
        self.assertEqual(oldmethod, fake_no_args)

    def test_patch_function_with_classmethod(self):
        oldmethod = fake_no_args
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.fake_no_args',
            D.bar_cls)
        with fixture:
            fake_no_args()
        self.assertEqual(oldmethod, fake_no_args)

    def test_patch_function_with_function(self):
        oldmethod = fake_no_args
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.fake_no_args',
            fake_no_args2)
        with fixture:
            fake_no_args()
        self.assertEqual(oldmethod, fake_no_args)

    def test_patch_function_with_partial(self):
        oldmethod = fake_no_args
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.fake_no_args',
            functools.partial(fake, 1))
        with fixture:
            (ret,) = fake_no_args()
            self.assertEqual(1, ret)
        self.assertEqual(oldmethod, fake_no_args)

    def test_patch_function_with_boundmethod(self):
        oldmethod = fake_no_args
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.fake_no_args',
            D().bar)
        with fixture:
            fake_no_args()
        self.assertEqual(oldmethod, fake_no_args)

    def test_patch_boundmethod_with_staticmethod(self):
        oldmethod = INST_C.foo
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.INST_C.foo',
            D.bar_static)
        with fixture:
            INST_C.foo()
        self.assertEqual(oldmethod, INST_C.foo)

    def test_patch_boundmethod_with_classmethod(self):
        oldmethod = INST_C.foo
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.INST_C.foo',
            D.bar_cls)
        with fixture:
            INST_C.foo()
        self.assertEqual(oldmethod, INST_C.foo)

    def test_patch_boundmethod_with_function(self):
        oldmethod = INST_C.foo
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.INST_C.foo',
            fake_no_args)
        with fixture:
            INST_C.foo()
        self.assertEqual(oldmethod, INST_C.foo)

    def test_patch_boundmethod_with_boundmethod(self):
        oldmethod = INST_C.foo
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.INST_C.foo',
            D().bar)
        with fixture:
            INST_C.foo()
        self.assertEqual(oldmethod, INST_C.foo)
        sentinel = object()
        self.assertEqual(sentinel, INST_C.__dict__.get('foo', sentinel))

    def test_patch_unboundmethod_with_staticmethod(self):
        oldmethod = C.foo
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo',
            D.bar_static_args)
        with fixture:
            target_self, arg = INST_C.foo(1)
            self.expectThat(target_self, Is(INST_C))
            self.assertEqual(1, arg)
        self.assertEqual(oldmethod, C.foo)

    def test_patch_unboundmethod_with_classmethod(self):
        oldmethod = C.foo
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo',
            D.bar_cls_args)
        with fixture:
            c = C()
            cls, target_self, arg = c.foo(1)
            self.expectThat(cls, Is(D))
            self.expectThat(target_self, Is(c))
            self.assertEqual(1, arg)
        self.assertEqual(oldmethod, C.foo)

    def test_patch_unboundmethod_with_function(self):
        oldmethod = C.foo
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo',
            fake)
        with fixture:
            c = C()
            target_self, arg = c.foo(1)
            self.expectThat(target_self, Is(c))
            self.assertTrue(1, arg)
        self.assertEqual(oldmethod, C.foo)

    def test_patch_unboundmethod_with_boundmethod(self):
        oldmethod = C.foo
        d = D()
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo',
            d.bar_two_args)
        with fixture:
            c = C()
            slf, target_self = c.foo()
            self.expectThat(slf, Is(d))
            self.expectThat(target_self, Is(c))
        self.assertEqual(oldmethod, C.foo)

    def test_double_patch_instancemethod(self):
        oldmethod = C.foo
        oldmethod_inst = C().foo
        fixture1 = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo',
            fake)
        fixture2 = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo',
            fake2)
        with fixture1:
            with fixture2:
                C().foo()
        self.assertEqual(oldmethod, C.foo)
        # The method address changes with each instantiation of C, and method
        # equivalence just tests that. Compare the code objects instead.
        self.assertEqual(oldmethod_inst.__code__, C().foo.__code__)

    def test_double_patch_staticmethod(self):
        oldmethod = C.foo_static
        oldmethod_inst = C().foo_static
        fixture1 = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_static',
            fake_no_args)
        fixture2 = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_static',
            fake_static)
        with fixture1:
            with fixture2:
                C.foo_static()
                C().foo_static()
        self._check_restored_static_or_class_method(oldmethod, oldmethod_inst,
            C, 'foo_static')

    def test_double_patch_classmethod(self):
        oldmethod = C.foo_cls
        oldmethod_inst = C().foo_cls
        fixture1 = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_cls',
            fake)
        fixture2 = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo_cls',
            fake2)
        with fixture1:
            with fixture2:
                C.foo_cls()
                C().foo_cls()
        self._check_restored_static_or_class_method(oldmethod, oldmethod_inst,
            C, 'foo_cls')

    def test_patch_c_foo_with_instance_d_bar_self_referential(self):
        oldmethod = C.foo
        oldmethod_inst = C().foo
        fixture = MonkeyPatch(
            'fixtures.tests._fixtures.test_monkeypatch.C.foo',
            D().bar_self_referential)
        with fixture:
            C().foo()
        self.assertEqual(oldmethod, C.foo)
        # The method address changes with each instantiation of C, and method
        # equivalence just tests that. Compare the code objects instead.
        self.assertEqual(oldmethod_inst.__code__, C().foo.__code__)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (C) 2011, Martin Pool <mbp@sourcefrog.net>
# 
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import signal
import time

import testtools
from testtools.testcase import (
    TestSkipped,
    )
from testtools.matchers import raises

import fixtures


def sample_timeout_passes():
    with fixtures.Timeout(100, gentle=True):
        pass  # Timeout shouldn't fire

def sample_long_delay_with_gentle_timeout():
    with fixtures.Timeout(1, gentle=True):
        time.sleep(100)  # Expected to be killed here.

def sample_long_delay_with_harsh_timeout():
    with fixtures.Timeout(1, gentle=False):
        time.sleep(100)  # Expected to be killed here.


class TestTimeout(testtools.TestCase, fixtures.TestWithFixtures):

    def requireUnix(self):
        if getattr(signal, 'alarm', None) is None:
            raise TestSkipped("no alarm() function")

    def test_timeout_passes(self):
        # This can pass even on Windows - the test is skipped.
        sample_timeout_passes()

    def test_timeout_gentle(self):
        self.requireUnix()
        self.assertRaises(
            fixtures.TimeoutException,
            sample_long_delay_with_gentle_timeout)

    def test_timeout_harsh(self):
        self.requireUnix()
        # This will normally kill the whole process, which would be
        # inconvenient.  Let's hook the alarm here so we can observe it.
        class GotAlarm(Exception):pass
        def sigalrm_handler(signum, frame):
            raise GotAlarm()
        old_handler = signal.signal(signal.SIGALRM, sigalrm_handler)
        self.addCleanup(signal.signal, signal.SIGALRM, old_handler)
        self.assertThat(sample_long_delay_with_harsh_timeout, raises(GotAlarm))



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2012, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

from testtools import TestCase
from testtools.matchers import Contains

from fixtures import (
    ByteStream,
    DetailStream,
    StringStream,
    )


class DetailStreamTest(TestCase):

    def test_doc_mentions_deprecated(self):
        self.assertThat(DetailStream.__doc__, Contains('Deprecated'))


class TestByteStreams(TestCase):

    def test_empty_detail_stream(self):
        detail_name = 'test'
        fixture = ByteStream(detail_name)
        with fixture:
            content = fixture.getDetails()[detail_name]
            self.assertEqual("", content.as_text())

    def test_stream_content_in_details(self):
        detail_name = 'test'
        fixture = ByteStream(detail_name)
        with fixture:
            stream = fixture.stream
            content = fixture.getDetails()[detail_name]
            # Output after getDetails is called is included.
            stream.write(b"testing 1 2 3")
            self.assertEqual("testing 1 2 3", content.as_text())

    def test_stream_content_reset(self):
        detail_name = 'test'
        fixture = ByteStream(detail_name)
        with fixture:
            stream = fixture.stream
            content = fixture.getDetails()[detail_name]
            stream.write(b"testing 1 2 3")
        with fixture:
            # The old content object returns the old usage
            self.assertEqual("testing 1 2 3", content.as_text())
            content = fixture.getDetails()[detail_name]
            # A new fixture returns the new output:
            stream = fixture.stream
            stream.write(b"1 2 3 testing")
            self.assertEqual("1 2 3 testing", content.as_text())


class TestStringStreams(TestCase):

    def test_empty_detail_stream(self):
        detail_name = 'test'
        fixture = StringStream(detail_name)
        with fixture:
            content = fixture.getDetails()[detail_name]
            self.assertEqual("", content.as_text())

    def test_stream_content_in_details(self):
        detail_name = 'test'
        fixture = StringStream(detail_name)
        with fixture:
            stream = fixture.stream
            content = fixture.getDetails()[detail_name]
            # Output after getDetails is called is included.
            stream.write("testing 1 2 3")
            self.assertEqual("testing 1 2 3", content.as_text())

    def test_stream_content_reset(self):
        detail_name = 'test'
        fixture = StringStream(detail_name)
        with fixture:
            stream = fixture.stream
            content = fixture.getDetails()[detail_name]
            stream.write("testing 1 2 3")
        with fixture:
            # The old content object returns the old usage
            self.assertEqual("testing 1 2 3", content.as_text())
            content = fixture.getDetails()[detail_name]
            # A new fixture returns the new output:
            stream = fixture.stream
            stream.write("1 2 3 testing")
            self.assertEqual("1 2 3 testing", content.as_text())



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

import warnings

import testtools

import fixtures


class TestWarningsCapture(testtools.TestCase, fixtures.TestWithFixtures):

    def test_capture_reuse(self):
        # DeprecationWarnings are hidden by default in Python 3.2+, enable them
        # https://docs.python.org/3/library/warnings.html#default-warning-filter
        self.useFixture(fixtures.WarningsFilter())
        warnings.simplefilter("always")

        w = fixtures.WarningsCapture()
        with w:
            warnings.warn("test", DeprecationWarning)
            self.assertEqual(1, len(w.captures))
        with w:
            self.assertEqual([], w.captures)

    def test_capture_message(self):
        # DeprecationWarnings are hidden by default in Python 3.2+, enable them
        # https://docs.python.org/3/library/warnings.html#default-warning-filter
        self.useFixture(fixtures.WarningsFilter())
        warnings.simplefilter("always")

        w = self.useFixture(fixtures.WarningsCapture())
        warnings.warn("hi", DeprecationWarning)
        self.assertEqual(1, len(w.captures))
        self.assertEqual("hi", str(w.captures[0].message))

    def test_capture_category(self):
        # DeprecationWarnings are hidden by default in Python 3.2+, enable them
        # https://docs.python.org/3/library/warnings.html#default-warning-filter
        self.useFixture(fixtures.WarningsFilter())
        warnings.simplefilter("always")

        w = self.useFixture(fixtures.WarningsCapture())
        categories = [
            DeprecationWarning, Warning, UserWarning,
            SyntaxWarning, RuntimeWarning,
            UnicodeWarning, FutureWarning,
        ]
        for category in categories:
            warnings.warn("test", category)
        self.assertEqual(len(categories), len(w.captures))
        for i, category in enumerate(categories):
            c = w.captures[i]
            self.assertEqual(category, c.category)


class TestWarningsFilter(testtools.TestCase, fixtures.TestWithFixtures):

    def test_filter(self):
        fixture = fixtures.WarningsFilter(
            [
                {
                    'action': 'ignore',
                    'category': DeprecationWarning,
                },
                {
                    'action': 'once',
                    'category': UserWarning,
                },
            ],
        )
        self.useFixture(fixture)
        with warnings.catch_warnings(record=True) as w:
            warnings.warn('deprecated', DeprecationWarning)
            warnings.warn('user', UserWarning)

        # only the user warning should be present, and it should only have been
        # raised once
        self.assertEqual(1, len(w))

    def test_filters_restored(self):

        class CustomWarning(Warning):
            pass

        fixture = fixtures.WarningsFilter(
            [
                {
                    'action': 'once',
                    'category': CustomWarning,
                },
            ],
        )

        # we copy the filter values rather than a reference to the containing
        # list since that can change
        old_filters = warnings.filters[:]

        # NOTE: we intentionally do not use 'self.useFixture' since we want to
        # teardown the fixture manually here before we exit this test method
        with fixture:
            new_filters = warnings.filters[:]
            self.assertEqual(len(old_filters) + 1, len(new_filters))
            self.assertNotEqual(old_filters, new_filters)

        new_filters = warnings.filters[:]
        self.assertEqual(len(old_filters), len(new_filters))
        self.assertEqual(old_filters, new_filters)
